from math import pi
from ngsolve import *
from netgen.geom2d import SplineGeometry
import numpy as np


u_exact = (sin(x)*sin(x)*cos(y)*sin(y),-cos(x)*sin(x)*sin(y)*sin(y))
p_exact = cos(x)*cos(y)
source = (-sin(x)*cos(y)-cos(x)*cos(x)*sin(2*y)+3*sin(x)*sin(x)*sin(2*y)
          ,-cos(x)*sin(y)+cos(y)*cos(y)*sin(2*x)-3*sin(y)*sin(y)*sin(2*x))

geo = SplineGeometry()
geo.AddRectangle((0,0),(pi,pi),bc="rectangle")
mesh = Mesh( geo.GenerateMesh(maxh = pi/10) )
order = 3
mesh.Curve(order)

print("ElementBoundary=",mesh.GetBoundaries())
print("NumEles=",mesh.ne)
print("NumEdges=",mesh.nedge)
print("NumVertex=",mesh.nv)

V = VectorH1(mesh,order=order, dirichlet="rectangle")
Q = H1(mesh,order=order-1)

X = FESpace([V,Q])

u,p = X.TrialFunction()
v,q = X.TestFunction()

print("V.ndof =", V.ndof,X.Range(0)) 
print("Q.ndof  =", Q.ndof,X.Range(1))
print("Total ndof=",X.ndof)

nu = 1.0 #fluid viscosity

stokes = nu*InnerProduct(grad(u), grad(v))+div(u)*q+div(v)*p - 1e-10*p*q
a = BilinearForm(X)
a += SymbolicBFI(stokes)
a.Assemble()

# nothing here ...
f = LinearForm(X)   
source = CoefficientFunction(source)
f += SymbolicLFI(-InnerProduct(source,v))
f.Assemble()

#Exact Dirichelt BC for velocity
gfu = GridFunction(X)

u_exact = CoefficientFunction(u_exact)
gfu.components[0].Set(u_exact, definedon=mesh.Boundaries("rectangle"))
print("Set Dirichelt velocity BC on ","rectangle")

inv_stokes = a.mat.Inverse(X.FreeDofs())

res = f.vec.CreateVector()
res.data = f.vec - a.mat*gfu.vec
gfu.vec.data += inv_stokes * res

Draw( gfu.components[0], mesh, "velocity" )
Draw( Norm(gfu.components[0]), mesh, "absvel(hdiv)")

vel = CoefficientFunction(gfu.components[0])
pres = CoefficientFunction(gfu.components[1])

#Error analysis
print ("Pres L2-error:", sqrt (Integrate ( (pres-p_exact)*(pres-p_exact), mesh)))
print ("Velocity L2-error:", sqrt (Integrate ( (vel-u_exact )*(vel-u_exact ), mesh)))

#Access solution 
PresDof=[i for i in V.Range(2)]
Pres_numpy=gfu.vec.FV().NumPy()[PresDof]
print("Min/Max Pressure on domain=(%lf,%lf)" %(min(Pres_numpy),max(Pres_numpy)))

#Output velocity and pressure to paraview
vtk = VTKOutput(ma=mesh,coefs=[vel,pres],names=["u","p"],filename="vtkout_CG",subdivision=1)
vtk.Do()
print("VTK file is generated!")